/*****************************************************************************/
/*                  Licensed Materials - Property of IBM                     */
/*                  (C) Copyright IBM Corp. 2000                             */
/*                                                                           */
/* Title           : MQSI v2.0x IMS Message Handler                          */
/*                   Plug-In Node                                            */
/*                                                                           */
/* Version         : 1.0 (25/07/2001 14:00)                                  */
/*                                                                           */
/* Program name    : IMSMSGHandler.c                                         */
/* Output name     : IMSMSGHandler.dll                                       */
/*                   (NOTE: must be renamed to IMSMSGHandler.lil)            */
/*                   <install dir>\bin for use with MQSI V2.0x broker        */
/*                                                                           */
/* Compile options : /VERBOSE /LD /MT ( PLG_DBG for user debugging )         */
/*                                                                           */
/* Link options    : /DLL                                                    */
/*                   LIB must contain <install dir>\lib\imbdfplg.lib         */
/*                                                                           */
/* Description     : MQSI v2.0x custom plug-in node that provides two basic  */
/*                   functions.                                              */
/*                   1) Takes Messages that contains an IMS header and an IMS*/
/*                      LLZZ segment/segments and strips out the LLZZ field  */
/*                      or fields and builds a message that contains only    */
/*                      the raw data.                                        */  
/*                   2) Takes IMS Messages that do not have an IMS header    */
/*                      and strips out the LLZZ field or fields and          */
/*                      builds a message that contains only the raw data     */
/*                                                                           */
/*                                                                           */
/*                   3) Takes a Message containing an IMS header and data    */
/*                      without and LLZZ fields and adds an LLZZ field to    */
/*                      create a single segment IMS message.                 */
/*                                                                           */
/*                   Customization of the Node functionality is provided by  */
/*                   the following attributes.                               */ 
/*                                                                           */
/*  Attribute      Initialized   Options   Action                            */
/*                                                                           */
/*1)MSG_TYPES  ANY/ALLIMS/IMSIIH ANY     - Messages without IIH propagated to*/
/*                                         the Output terminal               */  
/*                               IMSIIH  - Messages without IIH propagated to*/
/*                                         the failure terminal              */
/*                               ALLIMS  - Handle IMS Messages that may or   */
/*                                         may not have IMS headers          */ 
/*                                                                           */
/*2)DIRECTION      TOIMS/FROMIMS TOIMS   - LLZZ field added to make a single */
/*                                         segment IMS message               */          
/*                               FROMIMS - LLZZ field or fields stripped from*/
/*                                         IMS messages.                     */ 
/*3)MAXMSGDATASIZE 0000 bytes            - Maximum size of the message data  */
/*                                         this flow will handle excluding   */          
/*                                         the MD and IIH.                   */
/*4)DEBUG          NO/YES        NO      - No tracing                        */
/*                               YES     - Trace internal node processing to */ 
/*                                        C:\<install dir>IMSMSG.txt         */
/*                                                                           */
/* note:                                                                     */
/*      For performance debugging will only be available if the code has been*/ 
/*      built with the conditional compile option PLG_DBG. Add it to the     */
/*      Preprocessor Definitions in the C/C++ tag in settings if required    */
/*****************************************************************************/


/*****************************************************************************/
/* Include Headers List                                                      */
/*****************************************************************************/
#if defined (__CCI_SOLARIS__)||(__CCI_AIX__)
/* Unix specific headers                                                     */
#else
/* Windows specific headers                                                  */
 #include <windows.h>      /* Windows specific functions                     */
#endif 
/* Cross platform headers                                                    */
 #include <stdio.h>        /* Standard input/ouput functions                 */
 #include <stdlib.h>       /* Standard Library functions                     */
 #include <cmqc.h>         /* MQSeries application header                    */
 #include <BipCni.h>	   /* Function prototypes for Plug-in Node functions */
 #include <IMSMSGHandler.h>/* Specific header for this lil                   */

/*****************************************************************************/
/*****************************************************************************/
/* Node Utility Functions                                                    */
/*                                                                           */
/* insertAttributeTableEntry                                                 */
/* insertInputTerminalListEntry                                              */
/* insertOutputTerminalListEntry                                             */
/* getInputTerminalHandle                                                    */
/* getOutputTerminalHandle                                                   */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/

/*****************************************************************************/
/*****************************************************************************/
/* Node Implementation Functions                                             */
/*                                                                           */
/* These are mandatory and MUST be implemented by the programmer.            */
/*                                                                           */
/* createNodeContext                                                         */
/* deleteNodeContext                                                         */
/* getAttribute                                                              */
/* getAttributeName                                                          */
/* setAttribute                                                              */
/* bipGetMessageflowNodeFactory                                              */
/* evaluate                                                                  */
/*                                                                           */
/*****************************************************************************/
/*****************************************************************************/

/*****************************************************************************/
/* createNodeContext                                                         */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  factoryObject (input) - address of factory object that owns the node     */
/*                          being created.                                   */
/*  nodeName      (input) - name of node being created.                      */
/*  nodeObject    (input) - address of the node object that has just been    */
/*                          created.                                         */
/* Returns:                                                                  */
/*                                                                           */
/*  Pointer to the local context.                                            */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Creates any plug-in context for an instance of the node object. This is   */
/* invoked by the broker whenever a node instance is constructed. Nodes are  */
/* constructed whenever a messageflow containing the node is deployed by the */
/* broker.                                                                   */
/*                                                                           */
/* This function has 4 responsibilities:                                     */
/*                                                                           */
/* 1. Optionally verify the name of the node is supported by the factory.    */
/* 2. Allocate node instance specific data areas. e.g. For attribute data or */
/*    terminals.                                                             */
/* 3. Perform any additional resource acquisition or initialisation that is  */
/*    required for the processing of the node.                               */
/* 4. Return the address of the context to the calling function.             */
/*                                                                           */
/* Calls Functions:                                                          */
/* insertAttributeTableEntry                                                 */
/* insertInputTerminalListEntry                                              */
/* insertOutputTerminalListEntry                                             */
/*****************************************************************************/
void *createNodeContext(
						void *factoryObject
					   ,wchar_t *nodeName
					   ,void *nodeObject
					   )
{
  CONTEXT_STRUCTURE *p;

/* Allocate a pointer to the local context                                   */
  p = malloc(sizeof(CONTEXT_STRUCTURE));

/* If malloc was successful                                                  */
  if (p)
  {
    memset (p, 0, sizeof(CONTEXT_STRUCTURE)); /* Clear the context area      */
	p->nodeObject = nodeObject;
/*****************************************************************************/
/* Specify Node Name here                                                    */
/*****************************************************************************/
/* Verify the name of the node is supported by the factory                   */
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(nodeName, (CciChar*)constIMSNode) == 0) {
#else
	if (wcscmp(nodeName, L"IMSMSGHandlerNode") == 0)
#endif

	{
/* Build attribute list for the node                                         */
/*****************************************************************************/
/* Add Node attributes here                                                  */
/*****************************************************************************/
#if defined(__CCI_SOLARIS__)
    {
      const CciChar* ucsAttr = CciString("MSGTYPES") ;
      insertAttributeTableEntry(p, (CciChar*)ucsAttr, CCI_VALUE_TYPE_CHAR);
      free((void *)ucsAttr) ;
    }
    {
      const CciChar* ucsAttr = CciString("DIRECTION") ;
      insertAttributeTableEntry(p, (CciChar*)ucsAttr, CCI_VALUE_TYPE_CHAR);
      free((void *)ucsAttr) ;
    }

    {
      const CciChar* ucsAttr = CciString("MAXMSGDATASIZE") ;
      insertAttributeTableEntry(p, (CciChar*)ucsAttr, CCI_VALUE_TYPE_INTEGER);
      free((void *)ucsAttr) ;
    }
    {
      const CciChar* ucsAttr = CciString("DEBUG") ;
      insertAttributeTableEntry(p, (CciChar*)ucsAttr, CCI_VALUE_TYPE_CHAR);
      free((void *)ucsAttr) ;
    }
#else
	  insertAttributeTableEntry(p, nodeName, L"MSGTYPES", CCI_VALUE_TYPE_CHAR);
	  insertAttributeTableEntry(p, nodeName, L"DIRECTION", CCI_VALUE_TYPE_CHAR);
	  insertAttributeTableEntry(p, nodeName, L"MAXMSGDATASIZE", CCI_VALUE_TYPE_INTEGER);
	  insertAttributeTableEntry(p, nodeName, L"DEBUG", CCI_VALUE_TYPE_CHAR);
#endif

/*****************************************************************************/
/* Add Node terminals here                                                   */
/*****************************************************************************/
#if defined(__CCI_SOLARIS__)
    {
      const CciChar* ucsIn = CciString("in") ;
      insertInputTerminalListEntry(p, (CciChar*)ucsIn);
      free((void *)ucsIn) ;
    }
    {
      const CciChar* ucsOut = CciString("out") ;
      insertOutputTerminalListEntry(p, (CciChar*)ucsOut);
      free((void *)ucsOut) ;
    }
    {
      const CciChar* ucsFailure = CciString("failure") ;
      insertOutputTerminalListEntry(p, (CciChar*)ucsFailure);
      free((void *)ucsFailure) ;
    }
#else
	  insertInputTerminalListEntry(p, L"in");
	  insertOutputTerminalListEntry(p, L"out");
	  insertOutputTerminalListEntry(p, L"failure");
#endif

	}
	else
	{
/* Error handling if name of node is not supported by factory                */
	}
  }
  else
  {
/* Error handling if malloc failed                                           */
  }

/* Return the address of the context to the calling function                 */
  return(p);
}

/*****************************************************************************/
/* insertAttributeTableEntry                                                 */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context       (input) - pointer to the current context                   */
/*  nodeName      (input) - node name                                        */
/*  attributeName (input) - attribute name to be inserted into table         */
/*  type          (input) - type of the attribute                            */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  Pointer to an attribute table entry.                                     */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Acquires the storage for attribute and link it to the current context.    */
/*****************************************************************************/
ATTRIBUTE_TABLE_ENTRY *insertAttributeTableEntry(
												 CONTEXT_STRUCTURE *context
                                                ,wchar_t *nodeName
                                                ,wchar_t *attributeName
												,int type
                                                )
{
  ATTRIBUTE_TABLE_ENTRY *entry;
  int attributeSize;

/* Allocate storage for the attribute entry                                  */
  entry = malloc(sizeof(ATTRIBUTE_TABLE_ENTRY));

/* if malloc is successful                                                   */
  if (entry)
  {
/* set this entry to be the new end of list                                  */
    entry->next = 0;
/* store away the type for this entry                                        */
	entry->type = type;

/*****************************************************************************/
/* Add additional attributes types support here                              */
/*****************************************************************************/

	switch (entry->type)
	{
/* set up the size of the integer attributes                                 */
	  case CCI_VALUE_TYPE_INTEGER:
		attributeSize = 4;
		break;
/* set up the size of the character string attributes                        */
	  case CCI_VALUE_TYPE_CHAR:
		attributeSize = MAX_STRING_SIZE;
		break;
	  default:
		break;
	}
/* acquire and clear storage for attribute value                             */
	entry->valuePtr = malloc(attributeSize);
	if (entry->valuePtr)
	  memset(entry->valuePtr, 0, attributeSize);

/* store away the attribute name                                             */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, attributeName);
#else
	wcscpy (entry->name, attributeName);
#endif

/* if a previous element exists in the list, link it to this new one         */
	if (context->attributeListPrevious)
	  context->attributeListPrevious->next = entry;
	else if (context->attributeListHead == 0)
	  context->attributeListHead = entry;
	  
/* save the pointer to the previous element                                  */
	context->attributeListPrevious = entry;
  }
  return(entry);
}

/*****************************************************************************/
/* insertInputTerminalListEntry                                              */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context      (input) - pointer to the current context                    */
/*  terminalName (input) - name of terminal to be inserted into list         */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  Pointer to an input terminal list entry.                                 */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Acquires the storage for input terminal entry and link it to the current  */
/* context.                                                                  */
/*****************************************************************************/
TERMINAL_LIST_ENTRY *insertInputTerminalListEntry(
                                                  CONTEXT_STRUCTURE *context
											     ,wchar_t *terminalName 
                                                 )
{
  TERMINAL_LIST_ENTRY *entry;
  int rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry)
  {
/* This entry is the current end of the list                                 */
    entry->next = 0;

/* Store the terminal name                                                   */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, terminalName);
#else
    wcscpy(entry->name, terminalName);
#endif

/* Create terminal and save its handle                                       */
    entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, terminalName);

/* Link an existing previous element to this one                             */
    if (context->inputTerminalListPrevious)
	  context->inputTerminalListPrevious->next = entry;
    else if ((context->inputTerminalListHead) == 0)
	  context->inputTerminalListHead = entry;

/* Save the pointer to the previous element                                  */
    context->inputTerminalListPrevious = entry;
  }
  return(entry);
}

/*****************************************************************************/
/* insertOutputTerminalListEntry                                             */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  context      (input) - pointer to the current context                    */
/*  terminalName (input) - name of terminal to be inserted into list         */
/*                                                                           */
/*  Pointer to an output terminal list entry.                                */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Acquires the storage for output terminal entry and link it to the current */
/* context.                                                                  */
/*****************************************************************************/
TERMINAL_LIST_ENTRY *insertOutputTerminalListEntry(
                                                   CONTEXT_STRUCTURE *context
												  ,wchar_t *terminalName 
                                                  )
{
  TERMINAL_LIST_ENTRY *entry;
  int rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry)
  {
/* This entry is the current end of the list                                 */
    entry->next = 0;

/* Store the terminal name                                                   */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, terminalName);
#else
    wcscpy(entry->name, terminalName);
#endif

/* Create terminal and save its handle                                       */
    entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, terminalName);

/* Link an existing previous element to this one                             */
    if (context->outputTerminalListPrevious)
	  context->outputTerminalListPrevious->next = entry;
    else if ((context->outputTerminalListHead) == 0)
	  context->outputTerminalListHead = entry;

/* Save the pointer to the previous element                                  */
    context->outputTerminalListPrevious = entry;
  }
  return(entry);
}


/*****************************************************************************/
/* setAttribute                                                              */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context        (input)  - pointer to the current context                 */
/*  attributeName  (input)  - attribute name for which value is to be set    */
/*  attributeValue (input) -  value to be set for the attribute              */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  0 - success                                                              */
/*  1 - failure, unknown attribute name                                      */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Sets the value of an attribute on a specific node instance. It is invoked */
/* by the broker when a configuration request is received that attempts to   */
/* set the value of a node attribute. The broker will have verified that the */
/* attribute name is valid for the node.                                     */
/*                                                                           */
/* This function has 3 responsibilities:                                     */
/*                                                                           */
/* 1. Verify value of attribute is correctly specified, otherwise a          */
/*    configuration exception should be thrown.                              */
/* 2. Store the attribute value in the context.                              */
/* 3. Throw a configuration exception if errors occur.                       */
/*                                                                           */
/*****************************************************************************/

int setAttribute(
                 void *context
				,wchar_t *attributeName
				,wchar_t *attributeValue
                )
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;

  while (p != 0)
  {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, attributeName) == 0)
#else
    if (wcscmp(p->name, attributeName) == 0)
#endif
	{
/*****************************************************************************/
/* if additional attribute types required add support here                   */
/*****************************************************************************/
      switch(p->type)
	  {
/* copy your integer attributes based on their type size                     */
        case CCI_VALUE_TYPE_INTEGER:
#if defined(__CCI_SOLARIS__)
        {
          const char* mbAttrValue = mbString(attributeValue);
          *((int *)p->valuePtr) = atoi(mbAttrValue);
          free((void *)mbAttrValue);
        }
#elif defined(__CCI_AIX__)
          *((int *)p->valuePtr) = watoi(attributeValue);
#else
          *((int *)p->valuePtr) = _wtoi(attributeValue);
#endif
          break;
/* copy your character attributes based on their type size                   */
        case CCI_VALUE_TYPE_CHAR:
#if defined(__CCI_SOLARIS__)
        {
          const char* mbAttrValue = mbString(attributeValue);
          memcpy(p->valuePtr, mbAttrValue, MAX_STRING_SIZE);
          free((void *)mbAttrValue);
        }
#else
          memcpy(p->valuePtr, attributeValue, MAX_STRING_SIZE);
#endif
          break;

        default:
          break;
      }

/* Normal return                                                             */
      return(0);
    }
    p = p->next;
  }

/* Error...attribute not found                                               */
  return(1);
}

/*****************************************************************************/
/* getAttributeName                                                          */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context    (input)  - pointer to the current context                     */
/*  index      (input)  - index of the attribute for which the name is to be */
/*                        retrieved                                          */
/*  buffer     (output) - address of buffer into which attribute name is     */
/*                        copied                                             */
/*  bufferSize (input)  - length of buffer                                   */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  0 - success                                                              */
/*  1 - failure, unknown attribute name                                      */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Returns the name of an attribute by a given index. It is invoked by the   */
/* broker when it requires the names of attributes supported by a particular */
/* instance of a node. The function must gurantee to return the attribute    */
/* in a known, defined order and to return the attribute name represented by */
/* the index parameter.                                                      */
/*                                                                           */
/*****************************************************************************/
int getAttributeName(
                     void* context
					,int index
					,wchar_t* buffer
					,int bufferSize
                    )
{
  int count = 0;
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;

/* having set the pointer to the start of the attribute list, traverse down  */
/* and compare each element in the list until the indexed element is found   */
/* then return the name of the attribute in the supplied buffer.             */
  while (p != 0)
  {
    if (count == index)
	{
#if defined(__CCI_SOLARIS__)
      CciCharCpy(buffer, p->name);
#else
      wcscpy(buffer, p->name);
#endif
      return(0);
    }
    count++;
    p = p->next;
  }

/* Invalid index                                                             */
  return(1);
}

/*****************************************************************************/
/* bipGetMessageflowNodeFactory                                              */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  None.                                                                    */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  Address of factory object.                                               */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* This function is called when the message broker loads the "lil" during the*/
/* initialization of the broker engine                                       */
/*                                                                           */
/* This function has 5 responsibilities:                                     */
/*                                                                           */
/* 1. create the node factory (or factories) which the "lil" supports, using */
/*    the createNodeFactory() function, saving the returned pointer to the   */
/*    node factory for use at runtime.                                       */
/* 2. define each node supported by the (or each) factory) and pass a pointer*/
/*    to a function table which contains pointers to the node implementation */
/*    functions.                                                             */
/* 3. define the name of each attribute supported by each node, including a  */
/*    pointer to a validation function.                                      */
/* 4. define input and output terminals supported by each node.              */
/* 5. return the pointer to the factory, unless an error occurred, in which  */
/*    case zero should be returned.                                          */    
/*                                                                           */
/*****************************************************************************/
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
{
  static char* functionName="bipGetMessageflowNodeFactory()";
  void* factoryObject;
/*****************************************************************************/
/* Add node factory name here. Node name + Factory                           */
/*****************************************************************************/
#if defined(__CCI_SOLARIS__)
  CciChar factoryName[] = L"IMSMSGHandlerFactory";
#else
  wchar_t factoryName[] = L"IMSMSGHandlerFactory";
#endif

  CCI_EXCEPTION_ST exception_st;
#if defined(__CCI_SOLARIS__)
  /* Before we proceed we need to initialise all the static constants */
  initNodeConstants();
#endif

  int rc;
  static CNI_VFT vftable;

/* Setup virtual function table with addresses of implementation functions   */
  memset(&vftable, 0, sizeof(vftable));
  vftable.iFpCreateNodeContext     = createNodeContext;
  vftable.iFpGetAttributeName      = getAttributeName;
  vftable.iFpSetAttribute          = setAttribute;
  vftable.iFpGetAttribute          = getAttribute;
  vftable.iFpEvaluate              = evaluate;
/*****************************************************************************/
/* Customize log messages here                                               */
/*****************************************************************************/		
/* Output a log message                                                      */
  cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2",
         0, "Plugin <IMSMSGHandlerFactory> loading", (char *)0);

/* Create the Node Factory for this plug-in                                  */
#if defined(__CCI_SOLARIS__)
  factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPluginNodeFactory);
#else
  factoryObject = cniCreateNodeFactory(0, factoryName);
#endif


  if (factoryObject == CCI_NULL_ADDR)
  {
    if (rc == CCI_EXCEPTION)
	{
      cciGetLastExceptionData(&rc, &exception_st);
/* Any local error handling may go here                                      */
      cciRethrowLastException(&rc);
    }
/* Any further local error handling can go here                              */
  }
  else
  {
/*****************************************************************************/
/* Define node class type here                                               */
/*****************************************************************************/
/* Define a node type supported by our factory                               */
    cniDefineNodeClass(&rc, factoryObject, L"IMSMSGHandlerNode", &vftable); 
    if (rc != CCI_SUCCESS)
	{
/* Error Handling                                                            */
    }
  }
/* Return address of this factory object to the broker                       */
  return(factoryObject);
}


/*****************************************************************************/
/* evaluate                                                                  */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context         (input) - pointer to the current context                 */
/*  destinationList (input) - address of the input destination list for the  */
/*                            message                                        */
/*  exceptionList   (input) - address of the exception list for the message  */
/*  message         (input) - address of the input message object            */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  None.                                                                    */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Performs node processing. This function contains the main processing logic*/
/* of the message flow node. It is invoked by the broker when a message is   */
/* received on one of the input terminals of an instance of a node object.   */
/*                                                                           */
/* This function has 4 responsibilities:                                     */
/*                                                                           */
/* 1. Process the message in accordance with the values of any attribute on  */
/*    the node instance.                                                     */
/* 2. Propagate the message to the appropriate output terminal               */
/* 3. Throw an exception if errors occur.                                    */
/*                                                                           */
/* Calls Functions:                                                          */
/* getAttribute                                                              */
/* getOutputTerminalHandle                                                   */
/*****************************************************************************/
void evaluate(
              void* context
			 ,void* destinationList
			 ,void* exceptionList
			 ,void* message
             )
{
/* exception management fields                                               */
  CCI_EXCEPTION_ST exceptionStructure;
  void* outputTerminalObject;
   char functionName[] = "evaluate()"; 

/* cni function return codes                                                 */
  int rc;
  int retValue = 1;

/* debug and message support management fields                               */
  char DBGFLAG[3] = "YES";
  char IMSONLY[7] = "IMSONLY";
  char ANYDATA[3] = "ANY";
  char ALLIMS[6] =  "ALLIMS";
  char IMSIIH[6] =  "IMSIIH";
  char TO_IMS[5]  = "TOIMS";
  int DEBUGON = 0; /* debug defaults to off                                  */
  int ANYMSG = 0;  /* IMS only support or any msg                            */
  int IIH_REQD = 0;/* IMS IIH header required flag                           */
  int DIR = 0;     /* TOIMS or FROMIMS                                       */

/* message type and destination terminal and msg propagation indicators      */
  int MsgFormatType = 0;
  int DestinationTerminal = 0;
  int MessageToPropagate  = 0;

/* attribute management fields                                               */
  char DEBUG[MAX_STRING_SIZE]; 
  char MSGTYPES[MAX_STRING_SIZE]; 
  char DIRECTION[MAX_STRING_SIZE]; 
  int  MAXMSGDATASIZE = 0;
  int  attributeBufSize = 0;
  int  attributeLen = 0;
  wchar_t attributeBuffer[MAX_STRING_SIZE];

/* IMS segment management utility pointers                                   */
  char *pdest = 0;
  char *pdest2 =0;
/* IMS segment management numerics                                           */
  int TotalSegLen = 0;
  short int CurrSegLen = 0;
  byte SegLenHi = 0;
  byte SegLenLo = 0;
  int i;
  byte Z = 0;
  int FldLen = 0;

/* element value retrieval fields for byteArrays and character fields        */
  struct CciByteArray valueByteArray;
  CciSize valueByteArraySize = 0;
  wchar_t  valueCharFld[80];
  char  valueStrFld[80]; 

#if defined (PLG_DBG)
/* message buffer management fields for output in debug mode                 */
  CciSize bufferSize;
  CciByte bufferByte;
#endif

/* message and element pointers                                              */
  CciMessage*      outMsg;   
  CciMessage*      MsgToProp;

  CciElement*      MQMDHeaderChild;
  CciElement*      MQMDFormatChild;
  CciElement*      MQIIHHeaderChild;
  CciElement*      MQIIHFormatChild;
  CciElement*      IpBlobChild;
  CciElement*      IpBlobBlobChild;
  CciElement*      OpBlobChild;
  CciElement*      OpBlobBlobChild;
  CciElement*      inRootElement;
  CciElement*      outRootElement;

/*****************************************************************************/
/* Obtain and store all attribute values set for the node                    */
/*****************************************************************************/

 memset(DEBUG,0x00,MAX_STRING_SIZE);
 memset(MSGTYPES,0x00,MAX_STRING_SIZE);
 memset(DIRECTION,0x00,MAX_STRING_SIZE);

/* Get attribute value of "DEBUG_NODEBUG" attribute                          */

 rc = getAttribute(context, L"DEBUG", attributeBuffer, sizeof(attributeBuffer));
 pdest = (char *)&DEBUG; 
#if defined(__CCI_SOLARIS__)
 attributeLen = CciCharLen(attributeBuffer);
 pdest = mbString(attributeBuffer);
#else
 attributeLen = wcslen (attributeBuffer);
 wcstombs(pdest,attributeBuffer,attributeLen);
#endif

/* Get attribute value of "MSG_FORMATS" attribute                            */
 rc = getAttribute(context, L"MSGTYPES", attributeBuffer, sizeof(attributeBuffer));
 pdest = (char *)&MSGTYPES; 
#if defined(__CCI_SOLARIS__)
 attributeLen = CciCharLen(attributeBuffer);
 pdest = mbString(attributeBuffer);
#else
 attributeLen = wcslen (attributeBuffer);
 wcstombs(pdest,attributeBuffer,attributeLen);
#endif
 
/* Get attribute value of "DIRECTION" attribute                              */
 rc = getAttribute(context, L"DIRECTION", attributeBuffer, sizeof(attributeBuffer));
 pdest = (char *)&DIRECTION; 
#if defined(__CCI_SOLARIS__)
 attributeLen = CciCharLen(attributeBuffer);
 pdest = mbString(attributeBuffer);
#else
 attributeLen = wcslen (attributeBuffer);
 wcstombs(pdest,attributeBuffer,attributeLen);
#endif

/* Get attribute value of "MAXMSGDATASIZE" attribute */
 rc = getAttribute(context, L"MAXMSGDATASIZE", attributeBuffer, sizeof(attributeBuffer));
#if defined(__CCI_SOLARIS__)
 MAXMSGDATASIZE = _atoi(attributeBuffer);
#else
 MAXMSGDATASIZE = _wtoi(attributeBuffer);
#endif

/*****************************************************************************/
/* Determine if debug tracing to file has been selected                      */
/*****************************************************************************/
 rc = memcmp(DEBUG,DBGFLAG,sizeof(DBGFLAG));
 if (rc == 0){
/* open a file and dump the inbound message to a file                        */
#if defined (PLG_DBG)
    fptr = fopen("IMSMSG.txt", "a+");
#endif
    DEBUGON = 1;}

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
/* Get size of total message buffer                                          */
  bufferSize = cniBufferSize(&rc, message);

/* Loop for the length of bufferSize and copy the message into the file      */
  fprintf(fptr, "Message data in len = %d\n",bufferSize);
  fprintf(fptr, "Max Message data size = %d\n",MAXMSGDATASIZE);
  fprintf(fptr, "Complete message data in =\n ");
  for (i=0 ; i<=bufferSize ; i++)
  {
     bufferByte = cniBufferByte(&rc, message, i);
	 fprintf(fptr, "%c", (char*)bufferByte);
  }
  fprintf(fptr, "\n\n");
 } /*DEBUG test*/
#endif

/*****************************************************************************/
/* Determine if ANY messages are support or just IMS types                   */
/*****************************************************************************/
  rc = memcmp(MSGTYPES,ANYDATA,sizeof(ANYDATA)); /* compare with "ANY"       */
  if (rc == 0)
  {ANYMSG = 1;               /* support non IMS messages                     */
   IIH_REQD = 0;             /* an IIH not required                          */
#if defined (PLG_DBG)
   if (DEBUGON == 1)
   {fprintf(fptr, "\n Message format supported = ANY\n");}      /*DEBUG test */
#endif
  }else    
  {
   rc = memcmp(MSGTYPES,IMSIIH,sizeof(IMSIIH));  /* compare with "IMSIIH"    */
   if (rc == 0)
   {
   IIH_REQD = 1;             /* an IIH must be included                      */
   ANYMSG = 0;               /* do not support non IMS messages              */
#if defined (PLG_DBG)
     if (DEBUGON == 1)
	 {fprintf(fptr, "\n Message format supported = IMS with IIH only\n");} /*DEBUG test */
#endif
   }else
   {
    rc = memcmp(MSGTYPES,ALLIMS,sizeof(ALLIMS)); /* compare with "ALLIMS"    */ 
	if (rc ==0)
	{
    IIH_REQD = 0;             /* an IIH not required                         */
    ANYMSG = 0;               /* do not support non IMS messages             */
#if defined (PLG_DBG)
     if (DEBUGON == 1)
	 {fprintf(fptr, "\n Message format supported = All IMS types only\n");} /*DEBUG test */
#endif
	}else
	{ANYMSG = 1;               /* support non IMS messages                   */
     IIH_REQD = 0;             /* an IIH not required                        */
#if defined (PLG_DBG)
     if (DEBUGON == 1)
	 {fprintf(fptr, "\n Message format unknown = Support ANY types\n");}    /*DEBUG test */
#endif
	}
   }
  }

/*****************************************************************************/
/* Determine if DIRECTION of the flow. TOIMS or FROMIMS                      */
/*****************************************************************************/
  rc = memcmp(DIRECTION,TO_IMS,sizeof(TO_IMS));
  if (rc != 0)
  {DIR = FROMIMS;               /* support non IMS messages                  */
#if defined (PLG_DBG)
   if (DEBUGON == 1)
   {fprintf(fptr, "\n Message direction = FROMIMS\n");} /*DEBUG test         */
#endif
  }else
  {DIR = TOIMS;                /* support non IMS messages                   */
#if defined (PLG_DBG)
   if (DEBUGON == 1)
   {fprintf(fptr, "\n Message direction = TOIMS\n");}   /*DEBUG test         */
#endif
  }

/*****************************************************************************/
/* Initialize destination terminal to OUTPUT. And Msg to Propagate to OUTMSG */
/*****************************************************************************/
 DestinationTerminal = OUTPUT_TERM;
 MessageToPropagate  = OUTMSG;

/*****************************************************************************/
/* Address the input message.Create & address a msg object for the output msg*/
/*****************************************************************************/

  /**********************************************************/
  /* Get The Root Element Of The Input Message              */
  /**********************************************************/
  inRootElement = cniRootElement(&rc, message);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n inRootElement = cniRootElement(&rc, message) rc = %d\n",rc);} /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* Create A New Message For Output                        */
  /**********************************************************/
  outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message)) rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /*********************************************************/
  /* Get The Root Element Of The Output Message            */
  /*********************************************************/
  outRootElement = cniRootElement(&rc, outMsg);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n outRootElement = cniRootElement(&rc, outMsg) rc = %d\n",rc);} /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

/*****************************************************************************/
/* Manage the MQMD fields. Check the MD.Format value                         */
/*****************************************************************************/

  MQMDHeaderChild = cniSearchFirstChild(&rc, inRootElement,  CCI_COMPARE_MODE_NAME, 0, L"MQMD");
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n MQMDHeaderChild = cniSearchFirstChild(&rc, outRootElement,  CCI_COMPARE_MODE_NAME, 0, L'MQMD') rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  MQMDFormatChild = cniSearchFirstChild(&rc, MQMDHeaderChild, CCI_COMPARE_MODE_NAME, 0, L"Format");
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n MQMDFormatChild = cniSearchFirstChild(&rc, MQMDHeaderChild, CCI_COMPARE_MODE_NAME, 0, L'Format') rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* obtain element value                                   */
  /**********************************************************/
  memset(valueStrFld,0x20,sizeof(valueStrFld));
  memset(valueCharFld,0x20,sizeof(valueCharFld));

  valueByteArraySize = cniElementCharacterValue(&rc, MQMDFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld));
#if defined(__CCI_SOLARIS__)
  valueStrFld = mbString(valueCharFld);
#else
  wcstombs(valueStrFld,valueCharFld,sizeof(valueStrFld));
#endif

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n valueByteArraySize = cniElementCharacterValue(&rc, MQMDFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld)) rc = %d\n",rc);
  fprintf(fptr, "\n valueByteArraySize = cniElementCharacterValue(&rc, MQMDFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld)) data = %s\n",&valueStrFld);
  fprintf(fptr, "\n valueByteArraySize = cniElementCharacterValue(&rc, MQMDFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld)) valueByteArraySize = %d\n",valueByteArraySize);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* Check for MQFMT_IMS , MQFMT_IMS_VAR_STRING or OTHER    */
  /**********************************************************/
 FldLen = strlen(valueStrFld);
 rc = memcmp(valueStrFld,MQFMT_IMS,FldLen);
 if (rc == 0){
	 MsgFormatType = MSG_IMS_IIH;
 }else{
      rc = memcmp(valueStrFld,MQFMT_IMS_VAR_STRING,FldLen);
      if (rc == 0){
		  MsgFormatType = MSG_IMS_NO_IIH;
/* if the message hasn't got an IIH and the attribute states IMSIIH (IIH required)*/
/* propagate to failure                                                           */
		  if (IIH_REQD == 1)
		  {
          MessageToPropagate  = INMSG;
		  DestinationTerminal = FAILURE_TERM;
#if defined (PLG_DBG)
		  if (DEBUGON == 1){fprintf(fptr, "\n MsgFormatType = %d\n",MsgFormatType);} /*DEBUG test*/
#endif
          goto PROPAGATE_MSG;
          }

	  }else{
          /* not IMS data this may be INMSG to fail or INMSG to OUT */
          MessageToPropagate  = INMSG;
		  MsgFormatType = MSG_NO_IMS;
		  if (ANYMSG == 0)
		  {
		  DestinationTerminal = FAILURE_TERM;
          }else{
		  DestinationTerminal = OUTPUT_TERM;
		  }
#if defined (PLG_DBG)
		  if (DEBUGON == 1){fprintf(fptr, "\n MsgFormatType = %d\n",MsgFormatType);} /*DEBUG test*/
#endif
          goto PROPAGATE_MSG;
	  }
 }
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
    fprintf(fptr, "\n MsgFormatType = %d\n",MsgFormatType);} /*DEBUG test*/
#endif

/*********************************************************************************************/
/* Manage MQIIH fields. Check the IIH.Format it should be MQFMT_IMS_VAR_STRING               */
/*********************************************************************************************/
if (MsgFormatType == MSG_IMS_IIH)
{
  MQIIHHeaderChild = cniSearchFirstChild(&rc, inRootElement,  CCI_COMPARE_MODE_NAME, 0, L"MQIIH");
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n MQIIHHeaderChild = cniSearchFirstChild(&rc, outRootElement,  CCI_COMPARE_MODE_NAME, 0, L'MQIIH') rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  MQIIHFormatChild = cniSearchFirstChild(&rc, MQIIHHeaderChild, CCI_COMPARE_MODE_NAME, 0, L"Format");
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n MQIIHFormatChild = cniSearchFirstChild(&rc, MQIIHHeaderChild, CCI_COMPARE_MODE_NAME, 0, L'Format') rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* obtain element value if checking is necessary          */
  /**********************************************************/
  memset(valueStrFld,0x20,sizeof(valueStrFld));
  memset(valueCharFld,0x20,sizeof(valueCharFld));

  valueByteArraySize = cniElementCharacterValue(&rc, MQIIHFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld));

#if defined(__CCI_SOLARIS__)
  valueStrFld = mbString(valueCharFld);
#else
  wcstombs(valueStrFld,valueCharFld,sizeof(valueStrFld));
#endif

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n valueByteArraySize = cniElementCharacterValue(&rc, MQIIHFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld)) rc = %d\n",rc);
  fprintf(fptr, "\n valueByteArraySize = cniElementCharacterValue(&rc, MQIIHFormatChild, (CciChar*)&valueCharFld,sizeof(valueCharFld)) data = %s\n",&valueStrFld);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* MD states IMS msg + IIH. If no IIH or IIH.Format incorrect send to Failure Term*/
  /**********************************************************/
  FldLen = strlen(valueStrFld);
  rc = memcmp(valueStrFld,MQFMT_IMS_VAR_STRING,FldLen);
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

} /* endif MsgFormat Type */

/*****************************************************************************/
/* Manage the BLOB fields. Navigating to InputRoot.BLOB.BLOB for the data    */
/*****************************************************************************/
  /*********************************************************/
  /* copy the input message to the new output message      */
  /*********************************************************/

/* If you don't do this you can't navigate to inRootElement.BLOB.BLOB       */
/* The parser won't instantiate the unparsed elements unless you do this    */
  cniCopyElementTree(&rc, inRootElement, outRootElement);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n cniCopyElementTree(&rc, inRootElement, outRootElement) rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}
  /**********************************************************/
  /* navigate to OutputRoot."BLOB"                          */
  /**********************************************************/

  OpBlobChild = cniLastChild(&rc, outRootElement);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n OpBlobChild = cniLastChild(&rc, outRootElement) rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* navigate to OutputRoot."BLOB"."BLOB"                   */
  /**********************************************************/

  OpBlobBlobChild = cniLastChild(&rc, OpBlobChild);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n OpBlobBlobChild = cniLastChild(&rc, OpBlobChild)  rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* navigate to InputRoot."BLOB"                           */
  /**********************************************************/

  IpBlobChild = cniLastChild(&rc, inRootElement);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n IpBlobChild = cniLastChild(&rc, inRootElement) rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* navigate to InputRoot."BLOB"."BLOB"                    */
  /**********************************************************/

  IpBlobBlobChild = cniLastChild(&rc, IpBlobChild);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n IpBlobBlobChild = cniLastChild(&rc, IpBlobChild)  rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

  /**********************************************************/
  /* obtain element value. the IMS data in the blob         */
  /**********************************************************/
/*****************************************************************************/
/* Get size of total message buffer                                          */
/* bufferSize = cniBufferSize(&rc, message);                                 */
/* malloc size should be bufferSize - IIHlen - MDlen (optionally +4)         */
/* however the function seems unreliable and can return OK with a len of 0   */
/* If it worked we could dynamically malloc memory based on the msg data size*/
/* and we will be able to remove the MAXMSGDATASIZE attribute                */
/*****************************************************************************/

  valueByteArray.size = MAXMSGDATASIZE;
/* if we are to add an LLZZ up front we need an extra 4 bytes                */
  if (DIR == TOIMS){MAXMSGDATASIZE = MAXMSGDATASIZE+4;}

  valueByteArray.pointer = malloc(MAXMSGDATASIZE);
  pdest =  (char *)valueByteArray.pointer;          /* set utility ptr 1     */
  pdest2 = (char *)valueByteArray.pointer;          /* set utility ptr 2     */

/* if we are to add an LLZZ up front receive the buffer with a 4 byte gap    */
  if (DIR == TOIMS){pdest2 = pdest2+4;
	                valueByteArray.pointer = (void*)pdest2;}

  valueByteArraySize = cniElementByteArrayValue(&rc, IpBlobBlobChild, &valueByteArray);

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n valueByteArraySize = cniElementByteArrayValue(&rc, IpBlobBlobChild,valueByteArray) rc = %d\n",rc);
  pdest2 = pdest2+4;
  fprintf(fptr, "\n valueByteArraySize = cniElementByteArrayValue(&rc, IpBlobBlobChild,valueByteArray) data = %s\n",pdest2);
  pdest2 = pdest2-4;
  fprintf(fptr, "\n valueByteArray.size = %d\n",valueByteArray.size);
  fprintf(fptr, "\n valueByteArraySize = %d\n",valueByteArraySize);} /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}
if (DIR == FROMIMS)
{
/*****************************************************************************/
/* FROM IMS so need to strip LLZZ(s)                                         */
/*****************************************************************************/

TotalSegLen = 0;

/* pdest,pdest2,valueByteArray.pointer all pointing to head of malloc'd data */

while ( (pdest - (char *)valueByteArray.pointer)< valueByteArraySize)
{
  /**********************************************************/
  /* get the length from the LLZZ                           */
  /**********************************************************/
#if defined(__CCI_SOLARIS__)||(__CCI_AIX__)
/* big endian format length handling for unix                                */
   SegLenHi = *pdest;
   pdest++;
   SegLenLo  = *pdest;
   CurrSegLen = (SegLenLo*256 + SegLenHi) - 4; /* seg len less the llzz      */
/*************************/
/* from Barclays testing */
/*************************/
/* endian formatting for length handling                                     */
//   SegLenLo = *pdest;
//   pdest++;
//   SegLenHi  = *pdest;
//   CurrSegLen = (SegLenLo*256 + SegLenHi) - 4; /* seg len less the llzz      */

#else
/* little endian format length handling for intel                            */
   SegLenLo = *pdest;
   pdest++;
   SegLenHi  = *pdest;
   CurrSegLen = (SegLenLo*256 + SegLenHi) - 4; /* seg len less the llzz      */
/*************************/
/* from Barclays testing */
/*************************/
//   SegLenHi = *pdest;                                                        
//   pdest++;                                                                 
//   SegLenLo  = *pdest;                                                       
//   CurrSegLen = (SegLenLo*256 + SegLenHi) - 4;  /* seg len less the llzz      */

#endif
/* point pdest at the first byte after the llzz pdest2 points at llzz        */
   pdest = pdest + 3;
   i = 0;
   for (i=0 ; i<CurrSegLen ; i++)
   {
     *pdest2 = *pdest;
     pdest2++;
     pdest++;
   }
   TotalSegLen = TotalSegLen + CurrSegLen;
#if defined (PLG_DBG)
 if (DEBUGON == 1){
 fprintf(fptr, "\n Current Seg Len = %d\n",CurrSegLen);
 fprintf(fptr, " Total Seg Len   = %d\n",TotalSegLen);} /*DEBUG test*/
#endif
}
#if defined (PLG_DBG)
 if (DEBUGON == 1){
  fprintf(fptr, "\n Data After LLZZ stripped = %s\n",valueByteArray.pointer);
 } /*DEBUG test*/
#endif
  /***************************************************************************/
  /* copy the new buffer into OutputRoot blob.blob having adjusted the size  */
  /***************************************************************************/

  valueByteArray.size = TotalSegLen;
  cniSetElementByteArrayValue(&rc, OpBlobBlobChild, &valueByteArray);
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n cniSetElementByteArrayValue(&rc, OpBlobBlobChild, &valueByteArray); rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}
}else{ /* else if (DIR = FROMIMS) */

/*****************************************************************************/
/* TO IMS so need to add an LLZZ                                             */
/*****************************************************************************/
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
   fprintf(fptr, "\n Segment size without LLZZ = %d\n",valueByteArraySize);} /*DEBUG test*/
#endif

/* set up the CurrSegLen and point pdest2 at it                              */
   valueByteArray.size = valueByteArraySize + 4;
   CurrSegLen = valueByteArray.size;

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
   fprintf(fptr, "\n Segment size including LLZZ = %d\n",CurrSegLen);} /*DEBUG test*/
#endif
   pdest2 = (char*)&CurrSegLen;

#if defined(__CCI_SOLARIS__)||(__CCI_AIX__)
/* big endian type formatting of the LL for unix                             */
/* pdest1 is pointing to the very beginning of the malloc'd buffer           */
   *pdest = *pdest2; /* copy hi order length byte into the head of the buffer*/
    pdest++;
    pdest2++;
   *pdest = *pdest2; /* copy lo order length byte into the head of the buffer*/
    pdest++;
/*************************/
/* from barclays testing */
/*************************/
//  *pdest = *pdest2;    /*copy hi order length MQBYTE into the head of the buffer*/
//  pdest++;                                                                 
//  pdest2++;                                                                
//  *pdest = *pdest2;    /*copy lo order length MQBYTE into the head of the buffer*/
//  pdest++;                                                                 

#else
/* little endian type formatting of the LL for intel                         */
/* set pdest pointing to the very beginning of the malloc'd buffer +1        */
    pdest = pdest++;
   *pdest = *pdest2; /* copy hi order length byte into the head of the buffer*/
    pdest--;
	pdest2++;
   *pdest = *pdest2; /* copy lo order length byte into the head of the buffer*/
    pdest = pdest+2; /* move to ZZ place holder in the buffer                */
/*************************/
/* from barclays testing */
/*************************/
//    pdest = pdest++;
//   *pdest = *pdest2; /* copy hi order length MQBYTE into the head of the buffer*/
//    pdest--;
//	pdest2++;
//   *pdest = *pdest2; /* copy lo order length MQBYTE into the head of the buffer*/
//    pdest = pdest+2; /* move to ZZ place holder in the buffer                */

#endif
    pdest2 = (char *)&Z;
   *pdest = *pdest2; /* copy the 0x00,0x00 for the ZZ                        */
    pdest++;
   *pdest = *pdest2;

  /***************************************************************************/
  /* copy the new buffer into OutputRoot blob.blob having adjusted the size  */
  /* and set the pointer to the very beginning to include the new LLZZ       */ 
  /***************************************************************************/
  pdest2 = (char *)valueByteArray.pointer;
  pdest2 = pdest2 - 4;
  valueByteArray.pointer = (void *)pdest2;

  cniSetElementByteArrayValue(&rc, OpBlobBlobChild, &valueByteArray);

#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
  fprintf(fptr, "\n cniSetElementByteArrayValue(&rc, OpBlobBlobChild, &valueByteArray); rc = %d\n",rc);
 } /*DEBUG test*/
#endif
  if (rc != 0){ MessageToPropagate  = INMSG;
	            DestinationTerminal = FAILURE_TERM;
                goto PROPAGATE_MSG;}

} /* end if (DIR = FROMIMS) */

PROPAGATE_MSG:

 if(MessageToPropagate == INMSG)
 {     
#if defined (PLG_DBG)
	   if (DEBUGON == 1)  
       {fprintf(fptr, "\n Propagate Input Message\n");} /*DEBUG test         */
#endif      
	 MsgToProp = message;
 }else{
#if defined (PLG_DBG)
       if (DEBUGON == 1)
	   {fprintf(fptr, "\n Propagate Output Message\n");} /*DEBUG test        */
#endif
     MsgToProp = outMsg;
 }
#if defined (PLG_DBG)
 if (DEBUGON == 1){ 
/* Get size of total message buffer                                          */
  cniWriteBuffer(&rc, MsgToProp);
  bufferSize = cniBufferSize(&rc, MsgToProp);

/* Loop for the length of bufferSize and copy the message into the file      */
  fprintf(fptr, "\nMessage data out len = %d\n",bufferSize);
  fprintf(fptr, "Complete message data out = \n");
  i = 0;
  for (i=0 ; i<=bufferSize ; i++)
  {
     bufferByte = cniBufferByte(&rc, MsgToProp, i);
	 fprintf(fptr, "%c", (char*)bufferByte);
  }
  fprintf(fptr, "\n\n");
 } /*DEBUG test*/
 #endif

/*****************************************************************************/
/* propagate the message to the output or Failure terminal                   */
/*****************************************************************************/
 if (DestinationTerminal == OUTPUT_TERM)
 {
#if defined (PLG_DBG)
   if (DEBUGON == 1)
   {fprintf(fptr, "\n Propagate to out Terminal\n");}     /* DEBUG test      */
#endif
/* get the handle to the output terminal */
#if defined(__CCI_SOLARIS__)
   outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constOut);
#else
   outputTerminalObject = getOutputTerminalHandle(context, L"out");
#endif
 }else{
#if defined (PLG_DBG)
   if (DEBUGON == 1)
   {fprintf(fptr, "\n Propagate to failure Terminal\n");} /*DEBUG test       */
#endif
/* get the handle to the output terminal                                     */
#if defined(__CCI_SOLARIS__)
   outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constFailure);
#else
   outputTerminalObject = getOutputTerminalHandle(context, L"failure");
#endif
 }
/* close the output file                                                     */
#if defined (PLG_DBG)
  if (DEBUGON == 1){  fclose(fptr);} /*DEBUG test                            */
#endif
/* check to see if terminal exists and is attached                           */
  if (outputTerminalObject)
  {
	if (cniIsTerminalAttached(&rc, outputTerminalObject))
	{
	  if (rc == CCI_SUCCESS)
	  {
        retValue = cniPropagate(
			                    &rc
							   ,outputTerminalObject
							   ,destinationList
							   ,exceptionList
							   ,MsgToProp
							   );
		if (rc != CCI_SUCCESS)
		{
		  if (rc == CCI_EXCEPTION)
		  {
			cciGetLastExceptionData(&rc, &exceptionStructure);
			cciRethrowLastException(0);
		  }
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", BIP2623, "Message not propagated", (char *)0);
		}else{
		  cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", BIP2623, "Message propagated", (char *)0);
		}
	  }else{
        if (rc == CCI_EXCEPTION)
		{
          cciRethrowLastException(0);
		}else{
/* some other error, better log it                                           */
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", BIP2623, "Message not propagated", (char *)0);
        }
	  }
	}
  }else{
/* Terminal does not exist                                                   */
    cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2", BIP2623, "Terminal does not exist.", (char *)0);
  }
/* Delete the message we created now we have finished with it                */
  cniDeleteMessage(0, outMsg);
/* Free the malloc'd memory holding the message data                         */
  free(valueByteArray.pointer);
  return;
}
/*****************************************************************************/
/* getInputTerminalHandle                                                    */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context      (input) - pointer to the current context                    */
/*  terminalName (input) - name of terminal for handle to be returned        */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  The terminal object for a named input terminal.                          */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Traverses the input terminal list looking for the named input terminal.   */
/* When it finds the correct terminal, it returns the handle to it.          */
/*                                                                           */
/*****************************************************************************/
void *getInputTerminalHandle(
                             CONTEXT_STRUCTURE *context
							,wchar_t *terminalName
                            )
{
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->inputTerminalListHead;

/* having set the pointer to the start of the terminal list, traverse down   */
/* and compare each element in the list until the named terminal is found    */
/* then return the handle to that terminal object, otherwise return null     */
  while (p != 0)
  {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, terminalName) == 0) {
#else
    if (wcscmp(p->name, terminalName) == 0)
#endif
	  return(p->handle);
    p = p->next;
  }
/* No match was found                                                        */
  return(0);
}

/*****************************************************************************/
/* getOutputTerminalHandle                                                   */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context      (input) - pointer to the current context                    */
/*  terminalName (input) - name of terminal for handle to be returned        */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  The terminal object for a named output terminal.                         */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Traverses the output terminal list looking for the named output terminal. */
/* When it finds the correct terminal, it returns the handle to it.          */
/*                                                                           */
/*****************************************************************************/
void *getOutputTerminalHandle(
                              CONTEXT_STRUCTURE *context
			                 ,wchar_t *terminalName
                             )
{
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->outputTerminalListHead;

/* having set the pointer to the start of the terminal list, traverse down   */
/* and compare each element in the list until the named terminal is found    */
/* then return the handle to that terminal object, otherwise return null     */

  while (p != 0)
  {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, terminalName) == 0) {
#else
    if (wcscmp(p->name, terminalName) == 0)
#endif
	  return(p->handle);
    p = p->next;
  }
/* No match was found                                                        */
  return(0);
}


/*****************************************************************************/
/* getAttribute                                                              */
/*                                                                           */
/* Parameters:                                                               */
/*                                                                           */
/*  context       (input)  - pointer to the current context                  */
/*  attributeName (input)  - attribute name for which value is to be         */
/*                           retrieved                                       */
/*  buffer        (output) - address of buffer into which attribute value is */
/*                           copied                                          */
/*  bufferSize    (input)  - length of buffer                                */
/*                                                                           */
/* Returns:                                                                  */
/*                                                                           */
/*  0 - success                                                              */
/*  1 - failure, unknown attribute name                                      */
/*                                                                           */
/* Description:                                                              */
/*                                                                           */
/* Gets the value of an attribute on a specific node instance. It is invoked */
/* by the broker when a report request is received that causes a retrieval   */
/* of the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                     */
/*                                                                           */
/* This function has 3 responsibilities:                                     */
/*                                                                           */
/* 1. Return a character representation of the attribute value.              */
/* 2. Return a null string if  data is sensitive and should not be displayed.*/
/* 3. Throw an exception if errors occur.                                    */
/*                                                                           */
/*****************************************************************************/
int getAttribute(
                 void *context
				,wchar_t *attrName
				,wchar_t *buffer
				,int bufsize      
                )
{
  ATTRIBUTE_TABLE_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attributeListHead;
int i = 0;

#if defined(__CCI_SOLARIS__)
wchar_t              wstring[256];
char                  string[256];  
#endif

  while (p != 0)
  {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, attrName) == 0) {
#else
    if (wcscmp(p->name, attrName) == 0)
#endif
	{
/*****************************************************************************/
/* if additional attribute types required add support here                   */
/*****************************************************************************/
      switch(p->type)
	  {
        case CCI_VALUE_TYPE_INTEGER:
/* convert the integer type attribute value to string in buffer.             */
/* buffer points to AttributeBuffer                                          */
/* in evaluate function                                                      */
#if defined(__CCI_SOLARIS__)
        {
          sprintf((char*)&string, "%d", *((int*)p->valptr));
          const CciChar* ucsAttrValue = CciString(string);
          CciCharNCpy(buffer, ucsAttrValue, min(bufsize, 256));
          free((void *)ucsAttrValue);
        }
#endif
#if defined (__CCI_AIX__)
        sprintf((char*)&string, "%d", *((int*)p->valuePtr));
        mbstowcs((wchar_t*)&wstring, (char*)&string, 256);
		if (bufsize<256){
	        wcsncpy(buffer, (wchar_t*)&wstring, bufsize);
		}else{
            wcsncpy(buffer, (wchar_t*)&wstring, 256);
		}	
#endif
#if !defined __CCI_AIX__||__CCI_SOLARIS__
          _itow(*((int *)p->valuePtr), buffer, 10);
#endif



          break;
        case CCI_VALUE_TYPE_CHAR:
/* copy the character type attribute value to string in buffer.              */
/* buffer points to AttributeBuffer in evaluate function                     */
#if defined(__CCI_SOLARIS__)
        {
          const CciChar* ucsAttrValue = CciString(p->valptr);
          CciCharNCpy(buffer, ucsAttrValue, min(MAX_STRING_SIZE, bufsize));
          free((void *)ucsAttrValue);
        }
#else
			memcpy(buffer,p->valuePtr,min(MAX_STRING_SIZE,bufsize));
#endif
          break;
        default:
          break;
      }
/* Normal return                                                             */
      return(0);
    }
    p = p->next;
  }
/* Unknown attribute                                                         */
  return(1); 
}

#if defined(__CCI_SOLARIS__)
/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initNodeConstants(void)
{
  constZero = CciString(CONST_ZERO);
  constMQSeriesIntegrator2 = CciString(CONST_MQSERIES_INTEGRATOR_2);
  constSwitchTraceLocation = CciString(CONST_SWITCH_TRACE_LOCATION);
  constTransformTraceLocation = CciString(CONST_TRANSFORM_TRACE_LOCATION);
  constSwitchNode = CciString(CONST_SWITCH_NODE);
  constTransformNode = CciString(CONST_TRANSFORM_NODE);
  constPluginNodeFactory = CciString(CONST_PLUGIN_NODE_FACTORY);
  constNodeTraceOutfile = CciString(CONST_NODE_TRACE_OUT_FILE);
  constType = CciString(CONST_TYPE);
  constModify = CciString(CONST_MODIFY);
  constOut = CciString(CONST_OUT);
  constAdd = CciString(CONST_ADD);
  constChange = CciString(CONST_CHANGE);
  constDelete = CciString(CONST_DELETE);
  constHold = CciString(CONST_HOLD);
  constFailure = CciString(CONST_FAILURE);
}
#endif



